import ipaddress
from asyncio import CancelledError
from contextlib import suppress
from logging import getLogger

from defence360agent.contracts.messages import MessageType, Reject
from defence360agent.contracts.plugins import MessageSource, expect
from defence360agent.model.simplification import run_in_executor
from defence360agent.utils import recurring_check, timefun
from im360.api.ips import IPApi
from im360.contracts.plugins import IDSAwareMessageSink
from im360.ioc import services
from im360.internals import strategy
from im360.model.cache_sources import WhitelistCacheSourceFromCSF
from im360.model.firewall import IPList
from im360.subsys import csf
from im360.utils.net import unpack_ip_network

logger = getLogger(__name__)


class IgnoreWhitelisted(IDSAwareMessageSink):
    # NOTE: use IDSAwareMessageSink ^^ for IgnoreWhitelistedCSF
    PROCESSING_ORDER = IDSAwareMessageSink.ProcessingOrder.IGNORE_MESSAGE

    def __init__(self, cache=None):
        if cache is None:
            self._whitelist_cache = services.primary_whitelist_cache
        else:
            self._whitelist_cache = cache

    async def create_sink(self, loop):
        self._loop = loop

    async def process_one_message(self, message):
        if not message.get("attackers_ip"):
            return

        if await self._whitelist_cache.contains(message["attackers_ip"]):
            raise Reject("IP in the whitelist")

        return message

    @expect(MessageType.SensorAlert, MessageType.CaptchaDosAlert)
    async def process_alert(self, message):
        return await self.process_one_message(message)

    @expect(MessageType.SensorIncidentList)
    async def process_incident_list(self, incident_list):
        """
        Marks incidents from whitelisted IPs
        """
        for incident in incident_list.payload:
            if await self._whitelist_cache.contains(incident["attackers_ip"]):
                incident["ip_whitelisted"] = True

        return incident_list

    @expect(MessageType.SynclistResponse)
    @timefun(logger)
    async def process_synclist_response(self, message):
        """Process synclist response in batch."""
        assert isinstance(
            message["blocklist"], dict
        ), "blocklist should be dict type"

        subset_to_skip = await self._whitelist_cache.filter_contained(
            message["blocklist"].keys()
        )
        if subset_to_skip:
            logger.info(
                "Whitelisting %d ips on synclist response", len(subset_to_skip)
            )

        logger.debug(
            "Shrinking blocklist of size %d by %d element(s)...",
            len(message["blocklist"]),
            len(subset_to_skip),
        )

        for item in subset_to_skip:
            logger.debug("Shrinking blocklist by %s...", item)
            del message["blocklist"][item]

        logger.debug(
            "Shrinking blocklist by %d element(s)... done", len(subset_to_skip)
        )


class IgnoreWhitelistedCSF(IgnoreWhitelisted, MessageSource):
    STRATEGY = strategy.Strategy.CSF_COOP_STRATEGY

    def __init__(self, cache=None):
        if cache is None:
            self._whitelist_cache = services.csf_whitelist_cache
        else:
            self._whitelist_cache = cache

    async def create_sink(self, loop):
        self._loop = loop

    async def create_source(self, _, sink):
        # IgnoreWhitelistedCSF could generate UNBLOCK msg
        # from CSF's allow\ignore files, that why it becomes MessageSource
        self._sink = sink
        self._task = self._loop.create_task(self._sync_lists())

    async def activate(self):
        self.csf_whitelist = WhitelistCacheSourceFromCSF(
            *csf.CSF_IMUNIFY_IPLISTS_MAPPING["WHITE"]
        )
        return await super().activate()

    async def shutdown(self):
        self._task.cancel()
        with suppress(CancelledError):
            await self._task

    @recurring_check(120)
    async def _sync_lists(self):
        """
        Loads IPs from CSF conf files, unblock the ones
        that doesn't in cache and reloads cache
        """
        if not self.is_active():
            # not in CSF_COOP yet
            return
        all_ips = [ip for ip, comment in await self.csf_whitelist.fetch_all()]
        ips = await self._whitelist_cache.filter_not_contained(all_ips)
        if ips:
            logger.info(
                "Following IPs not in whitelist CSF-cache %s,syncing them..."
                % ips
            )

            for ip in ips:
                await self._unblock_whitelisted(ip)
            # rebuilding cache with all new IPs
            self._whitelist_cache.reset()

    async def _unblock_whitelisted(self, ip):
        """Given *ip* (net) from CSF's white/ignore lists that is not yet in
        the agent's whitelist cache,
        unblock the net from GRAY/SPLASH lists.

        If there is a supernet (containing the net), put the net into
        agent's ignore list.

        """
        net = ipaddress.ip_network(ip)

        # unblock the ip from the lists
        for listname in [IPList.BLACK, IPList.GRAY, IPList.GRAY_SPLASHSCREEN]:
            # find supernets that contain ip
            supernets = await run_in_executor(
                self._loop,
                lambda: IPList.find_closest_ip_nets(
                    net, [listname], manual=False
                ),
            )
            # note: supernets are sorted from smallest to largest (>)
            for iplist in supernets:
                supernet = iplist.ip_network
                if supernet == net:  # can unblock
                    affected, not_affected = await IPApi.unblock(
                        items=[supernet],
                        listname=listname,
                    )
                    if not_affected:
                        logger.warning(
                            "Can't unblock following IPs from %s list: %s",
                            listname,
                            not_affected,
                        )  # pragma: no cover
                    logger.info(
                        "Following IPs were unblocked from %s list: %s",
                        listname,
                        affected,
                    )
                else:  # can't unblock, put the net into the ignore list
                    assert supernet < net  # larger network is less
                    # we have a larger network in the list
                    # no way to unblock the ip, because it is a part of
                    # blocked supernet so, sending ClientUnblock
                    # in order to put the ip into ignore list
                    await self._sink.process_message(
                        MessageType.ClientUnblock(
                            attackers_ip=net, plugin_id="imunify360"
                        )
                    )
                    # no point to continue:
                    #  the ignore list applies to all lists
                    return
